@@ -66,20 +66,20 @@ static struct nullfs_fsal_obj_handle *nullfs_alloc_handle(
 		struct fsal_obj_handle *sub_handle,
 		struct fsal_filesystem *fs)
 {
-	struct nullfs_fsal_obj_handle *result =
-		gsh_calloc(1, sizeof(struct nullfs_fsal_obj_handle));
-	if (result) {
-		/* attributes */
-		result->obj_handle.attrs = sub_handle->attrs;
-		/* default handlers */
-		fsal_obj_handle_init(&result->obj_handle, &export->export,
-				     sub_handle->type);
-		/* nullfs handlers */
-		nullfs_handle_ops_init(&result->obj_handle.obj_ops);
-		result->sub_handle = sub_handle;
-		result->obj_handle.type = sub_handle->type;
-		result->obj_handle.fs = fs;
-	}
+	struct nullfs_fsal_obj_handle *result;
+
+	result = gsh_calloc(1, sizeof(struct nullfs_fsal_obj_handle));
+
+	/* default handlers */
+	fsal_obj_handle_init(&result->obj_handle, &export->export,
+			     sub_handle->type);
+	/* nullfs handlers */
+	nullfs_handle_ops_init(&result->obj_handle.obj_ops);
+	result->sub_handle = sub_handle;
+	result->obj_handle.type = sub_handle->type;
+	result->obj_handle.fsid = sub_handle->fsid;
+	result->obj_handle.fileid = sub_handle->fileid;
+	result->obj_handle.fs = fs;
 
 	return result;
 }
@@ -110,16 +110,11 @@ static fsal_status_t nullfs_alloc_and_check_handle(
 	fsal_status_t status = subfsal_status;
 
 	if (!FSAL_IS_ERROR(subfsal_status)) {
-		struct nullfs_fsal_obj_handle *null_handle =
-			nullfs_alloc_handle(export, sub_handle, fs);
-		if (null_handle == NULL) {
-			status = fsalstat(ERR_FSAL_NOMEM, ENOMEM);
-			LogCrit(COMPONENT_FSAL, "Out of memory");
+		struct nullfs_fsal_obj_handle *null_handle;
 
-			sub_handle->obj_ops.release(sub_handle);
-		} else {
-			*new_handle = &null_handle->obj_handle;
-		}
+		null_handle = nullfs_alloc_handle(export, sub_handle, fs);
+
+		*new_handle = &null_handle->obj_handle;
 	}
 	return status;
 }
@@ -129,7 +124,8 @@ static fsal_status_t nullfs_alloc_and_check_handle(
  */
 
 static fsal_status_t lookup(struct fsal_obj_handle *parent,
-			    const char *path, struct fsal_obj_handle **handle)
+			    const char *path, struct fsal_obj_handle **handle,
+			    struct attrlist *attrs_out)
 {
 	/** Parent as nullfs handle.*/
 	struct nullfs_fsal_obj_handle *null_parent =
@@ -146,9 +142,9 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 	struct nullfs_fsal_export *export =
 		container_of(op_ctx->fsal_export, struct nullfs_fsal_export,
 			     export);
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	status = null_parent->sub_handle->obj_ops.lookup(
-			null_parent->sub_handle, path, &sub_handle);
+			null_parent->sub_handle, path, &sub_handle, attrs_out);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
@@ -157,8 +153,9 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 }
 
 static fsal_status_t create(struct fsal_obj_handle *dir_hdl,
-			    const char *name, struct attrlist *attrib,
-			    struct fsal_obj_handle **handle)
+			    const char *name, struct attrlist *attrs_in,
+			    struct fsal_obj_handle **new_obj,
+			    struct attrlist *attrs_out)
 {
 	/** Parent directory nullfs handle. */
 	struct nullfs_fsal_obj_handle *nullfs_dir =
@@ -172,24 +169,25 @@ static fsal_status_t create(struct fsal_obj_handle *dir_hdl,
 	/** Subfsal handle of the new file.*/
 	struct fsal_obj_handle *sub_handle;
 
-	*handle = NULL;
+	*new_obj = NULL;
 
 	/* creating the file with a subfsal handle. */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = nullfs_dir->sub_handle->obj_ops.create(
-		nullfs_dir->sub_handle, name, attrib, &sub_handle);
+		nullfs_dir->sub_handle, name, attrs_in, &sub_handle, attrs_out);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
 	return nullfs_alloc_and_check_handle(export, sub_handle, dir_hdl->fs,
-					     handle, status);
+					     new_obj, status);
 }
 
 static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
-			     const char *name, struct attrlist *attrib,
-			     struct fsal_obj_handle **handle)
+			     const char *name, struct attrlist *attrs_in,
+			     struct fsal_obj_handle **new_obj,
+			     struct attrlist *attrs_out)
 {
-	*handle = NULL;
+	*new_obj = NULL;
 	/** Parent directory nullfs handle. */
 	struct nullfs_fsal_obj_handle *parent_hdl =
 		container_of(dir_hdl, struct nullfs_fsal_obj_handle,
@@ -203,21 +201,23 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 	struct fsal_obj_handle *sub_handle;
 
 	/* Creating the directory with a subfsal handle. */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = parent_hdl->sub_handle->obj_ops.mkdir(
-		parent_hdl->sub_handle, name, attrib, &sub_handle);
+		parent_hdl->sub_handle, name, attrs_in, &sub_handle, attrs_out);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
 	return nullfs_alloc_and_check_handle(export, sub_handle, dir_hdl->fs,
-					     handle, status);
+					     new_obj, status);
 }
 
 static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
-			      const char *name, object_file_type_t nodetype,
-			      fsal_dev_t *dev,	/* IN */
-			      struct attrlist *attrib,
-			      struct fsal_obj_handle **handle)
+			      const char *name,
+			      object_file_type_t nodetype,
+			      fsal_dev_t *dev,
+			      struct attrlist *attrs_in,
+			      struct fsal_obj_handle **new_obj,
+			      struct attrlist *attrs_out)
 {
 	/** Parent directory nullfs handle. */
 	struct nullfs_fsal_obj_handle *nullfs_dir =
@@ -231,18 +231,18 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 	/** Subfsal handle of the new node.*/
 	struct fsal_obj_handle *sub_handle;
 
-	*handle = NULL;
+	*new_obj = NULL;
 
 	/* Creating the node with a subfsal handle. */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = nullfs_dir->sub_handle->obj_ops.mknode(
-		nullfs_dir->sub_handle, name, nodetype, dev, attrib,
-		&sub_handle);
+		nullfs_dir->sub_handle, name, nodetype, dev, attrs_in,
+		&sub_handle, attrs_out);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
 	return nullfs_alloc_and_check_handle(export, sub_handle, dir_hdl->fs,
-					     handle, status);
+					     new_obj, status);
 }
 
 /** makesymlink
@@ -252,9 +252,11 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
  */
 
 static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
-				 const char *name, const char *link_path,
-				 struct attrlist *attrib,
-				 struct fsal_obj_handle **handle)
+				 const char *name,
+				 const char *link_path,
+				 struct attrlist *attrs_in,
+				 struct fsal_obj_handle **new_obj,
+				 struct attrlist *attrs_out)
 {
 	/** Parent directory nullfs handle. */
 	struct nullfs_fsal_obj_handle *nullfs_dir =
@@ -268,17 +270,18 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 	/** Subfsal handle of the new link.*/
 	struct fsal_obj_handle *sub_handle;
 
-	*handle = NULL;
+	*new_obj = NULL;
 
 	/* creating the file with a subfsal handle. */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = nullfs_dir->sub_handle->obj_ops.symlink(
-		nullfs_dir->sub_handle, name, link_path, attrib, &sub_handle);
+		nullfs_dir->sub_handle, name, link_path, attrs_in, &sub_handle,
+		attrs_out);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
 	return nullfs_alloc_and_check_handle(export, sub_handle, dir_hdl->fs,
-					     handle, status);
+					     new_obj, status);
 }
 
 static fsal_status_t readsymlink(struct fsal_obj_handle *obj_hdl,
@@ -292,7 +295,7 @@ static fsal_status_t readsymlink(struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status =
 		handle->sub_handle->obj_ops.readlink(handle->sub_handle,
 						     link_content, refresh);
@@ -314,7 +317,7 @@ static fsal_status_t linkfile(struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = handle->sub_handle->obj_ops.link(
 		handle->sub_handle, nullfs_dir->sub_handle, name);
 	op_ctx->fsal_export = &export->export;
@@ -335,16 +338,17 @@ static fsal_status_t linkfile(struct fsal_obj_handle *obj_hdl,
  *
  * @return Result coming from the upper layer.
  */
-static bool nullfs_readdir_cb(const char *name, void *dir_state,
-			       fsal_cookie_t cookie)
+static bool nullfs_readdir_cb(const char *name, struct fsal_obj_handle *obj,
+			      struct attrlist *attrs,
+			      void *dir_state, fsal_cookie_t cookie)
 {
 	struct nullfs_readdir_state *state =
 		(struct nullfs_readdir_state *) dir_state;
 
 	op_ctx->fsal_export = &state->exp->export;
-	bool result = state->cb(name, state->dir_state, cookie);
+	bool result = state->cb(name, obj, attrs, state->dir_state, cookie);
 
-	op_ctx->fsal_export = state->exp->sub_export;
+	op_ctx->fsal_export = state->exp->export.sub_export;
 
 	return result;
 }
@@ -362,7 +366,8 @@ static bool nullfs_readdir_cb(const char *name, void *dir_state,
 
 static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 				  fsal_cookie_t *whence, void *dir_state,
-				  fsal_readdir_cb cb, bool *eof)
+				  fsal_readdir_cb cb, attrmask_t attrmask,
+				  bool *eof)
 {
 	struct nullfs_fsal_obj_handle *handle =
 		container_of(dir_hdl, struct nullfs_fsal_obj_handle,
@@ -379,10 +384,10 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 	};
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status =
 		handle->sub_handle->obj_ops.readdir(handle->sub_handle,
-		whence, &cb_state, nullfs_readdir_cb, eof);
+		whence, &cb_state, nullfs_readdir_cb, attrmask, eof);
 	op_ctx->fsal_export = &export->export;
 
 	return status;
@@ -409,7 +414,7 @@ static fsal_status_t renamefile(struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = nullfs_olddir->sub_handle->obj_ops.rename(
 		nullfs_obj->sub_handle, nullfs_olddir->sub_handle,
 		old_name, nullfs_newdir->sub_handle, new_name);
@@ -418,7 +423,8 @@ static fsal_status_t renamefile(struct fsal_obj_handle *obj_hdl,
 	return status;
 }
 
-static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl)
+static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
+			      struct attrlist *attrib_get)
 {
 	struct nullfs_fsal_obj_handle *handle =
 		container_of(obj_hdl, struct nullfs_fsal_obj_handle,
@@ -429,9 +435,10 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl)
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status =
-		handle->sub_handle->obj_ops.getattrs(handle->sub_handle);
+		handle->sub_handle->obj_ops.getattrs(handle->sub_handle,
+						     attrib_get);
 	op_ctx->fsal_export = &export->export;
 
 	return status;
@@ -454,7 +461,7 @@ static fsal_status_t setattrs(struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = handle->sub_handle->obj_ops.setattrs(
 		handle->sub_handle, attrs);
 	op_ctx->fsal_export = &export->export;
@@ -467,19 +474,23 @@ static fsal_status_t setattrs(struct fsal_obj_handle *obj_hdl,
  */
 
 static fsal_status_t file_unlink(struct fsal_obj_handle *dir_hdl,
+				 struct fsal_obj_handle *obj_hdl,
 				 const char *name)
 {
 	struct nullfs_fsal_obj_handle *nullfs_dir =
 		container_of(dir_hdl, struct nullfs_fsal_obj_handle,
 			     obj_handle);
+	struct nullfs_fsal_obj_handle *nullfs_obj =
+		container_of(obj_hdl, struct nullfs_fsal_obj_handle,
+			     obj_handle);
 	struct nullfs_fsal_export *export =
 		container_of(op_ctx->fsal_export, struct nullfs_fsal_export,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = nullfs_dir->sub_handle->obj_ops.unlink(
-		nullfs_dir->sub_handle, name);
+		nullfs_dir->sub_handle, nullfs_obj->sub_handle, name);
 	op_ctx->fsal_export = &export->export;
 
 	return status;
@@ -505,7 +516,7 @@ static fsal_status_t handle_digest(const struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	fsal_status_t status = handle->sub_handle->obj_ops.handle_digest(
 		handle->sub_handle, output_type, fh_desc);
 	op_ctx->fsal_export = &export->export;
@@ -532,7 +543,7 @@ static void handle_to_key(struct fsal_obj_handle *obj_hdl,
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	handle->sub_handle->obj_ops.handle_to_key(handle->sub_handle, fh_desc);
 	op_ctx->fsal_export = &export->export;
 }
@@ -553,7 +564,7 @@ static void release(struct fsal_obj_handle *obj_hdl)
 			     export);
 
 	/* calling subfsal method */
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
 	hdl->sub_handle->obj_ops.release(hdl->sub_handle);
 	op_ctx->fsal_export = &export->export;
 
@@ -572,7 +583,6 @@ void nullfs_handle_ops_init(struct fsal_obj_ops *ops)
 	ops->mknode = makenode;
 	ops->symlink = makesymlink;
 	ops->readlink = readsymlink;
-	ops->test_access = fsal_test_access;
 	ops->getattrs = getattrs;
 	ops->setattrs = setattrs;
 	ops->link = linkfile;
@@ -585,7 +595,6 @@ void nullfs_handle_ops_init(struct fsal_obj_ops *ops)
 	ops->commit = nullfs_commit;
 	ops->lock_op = nullfs_lock_op;
 	ops->close = nullfs_close;
-	ops->lru_cleanup = nullfs_lru_cleanup;
 	ops->handle_digest = handle_digest;
 	ops->handle_to_key = handle_to_key;
 
@@ -596,7 +605,6 @@ void nullfs_handle_ops_init(struct fsal_obj_ops *ops)
 	ops->getextattr_value_by_id = nullfs_getextattr_value_by_id;
 	ops->setextattr_value = nullfs_setextattr_value;
 	ops->setextattr_value_by_id = nullfs_setextattr_value_by_id;
-	ops->getextattr_attrs = nullfs_getextattr_attrs;
 	ops->remove_extattr_by_id = nullfs_remove_extattr_by_id;
 	ops->remove_extattr_by_name = nullfs_remove_extattr_by_name;
 
@@ -612,7 +620,8 @@ void nullfs_handle_ops_init(struct fsal_obj_ops *ops)
 
 fsal_status_t nullfs_lookup_path(struct fsal_export *exp_hdl,
 				 const char *path,
-				 struct fsal_obj_handle **handle)
+				 struct fsal_obj_handle **handle,
+				 struct attrlist *attrs_out)
 {
 	/** Handle given by the subfsal. */
 	struct fsal_obj_handle *sub_handle = NULL;
@@ -625,9 +634,12 @@ fsal_status_t nullfs_lookup_path(struct fsal_export *exp_hdl,
 	/* call to subfsal lookup with the good context. */
 	fsal_status_t status;
 
-	op_ctx->fsal_export = exp->sub_export;
-	status = exp->sub_export->exp_ops.lookup_path(exp->sub_export, path,
-						      &sub_handle);
+	op_ctx->fsal_export = exp->export.sub_export;
+
+	status = exp->export.sub_export->exp_ops.lookup_path(
+				exp->export.sub_export, path, &sub_handle,
+				attrs_out);
+
 	op_ctx->fsal_export = &exp->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
@@ -650,7 +662,8 @@ fsal_status_t nullfs_lookup_path(struct fsal_export *exp_hdl,
 
 fsal_status_t nullfs_create_handle(struct fsal_export *exp_hdl,
 				   struct gsh_buffdesc *hdl_desc,
-				   struct fsal_obj_handle **handle)
+				   struct fsal_obj_handle **handle,
+				   struct attrlist *attrs_out)
 {
 	/** Current nullfs export. */
 	struct nullfs_fsal_export *export =
@@ -662,10 +675,12 @@ fsal_status_t nullfs_create_handle(struct fsal_export *exp_hdl,
 	/* call to subfsal lookup with the good context. */
 	fsal_status_t status;
 
-	op_ctx->fsal_export = export->sub_export;
+	op_ctx->fsal_export = export->export.sub_export;
+
+	status = export->export.sub_export->exp_ops.create_handle(
+			export->export.sub_export, hdl_desc, &sub_handle,
+			attrs_out);
 
-	status = export->sub_export->exp_ops.create_handle(export->sub_export,
-		hdl_desc, &sub_handle);
 	op_ctx->fsal_export = &export->export;
 
 	/* wraping the subfsal handle in a nullfs handle. */
